from typing import Callable

import catalogue
import pytest

import xpersist


@pytest.mark.parametrize(
    'registry_name,func_name,expected',
    [
        ('serializers', 'xarray.netcdf', True),
        ('serializers', 'my_func', False),
        ('my_registry', 'test', False),
    ],
)
def test_has(registry_name, func_name, expected):
    assert xpersist.registry.has(registry_name, func_name) == expected


@pytest.mark.parametrize(
    'registry_name,func_name',
    [
        ('serializers', 'xarray.netcdf'),
        ('serializers', 'xarray.zarr'),
        ('serializers', 'joblib'),
        ('serializers', 'xcollection'),
    ],
)
def test_get(registry_name, func_name):
    assert isinstance(xpersist.registry.get(registry_name, func_name), Callable)


def test_get_error():
    with pytest.raises(ValueError):
        xpersist.registry.get('my_registry', 'my_func')

    with pytest.raises(catalogue.RegistryError):
        xpersist.registry.get('serializers', 'my_func')


def test_create_error():
    with pytest.raises(ValueError):
        xpersist.registry.create('serializers')


def test_create():
    xpersist.registry.create('my_registry')

    @xpersist.registry.my_registry.register('test')
    def my_func():
        return

    assert xpersist.registry.has('my_registry', 'test')



import os
import time

import pytest
import xarray as xr
from prefect import Flow, task
from prefect.executors import DaskExecutor, LocalExecutor

from xpersist import CacheStore
from xpersist.prefect.result import XpersistResult

ds = xr.tutorial.open_dataset('air_temperature').isel(time=0)


@pytest.mark.parametrize(
    'data, serializer',
    [
        ([1, 2, 3], 'auto'),
        ({'foo': 1, 'bar': 2}, 'auto'),
        (ds, 'xarray.netcdf'),
        ((ds, 'xarray.zarr')),
    ],
)
def test_result(data, serializer):
    r = XpersistResult(
        CacheStore(),
        serializer=serializer,
    )

    new = r.write(data)
    assert new.read(new.location).value == data
    assert r.cache_store.get(new.location) == data
    assert r.exists(new.location)


@pytest.mark.parametrize(
    'executor',
    [
        LocalExecutor(),
        DaskExecutor(cluster_kwargs={'processes': False, 'threads_per_worker': 8}, debug=True),
    ],
)
def test_result_flow(executor):
    os.environ['PREFECT__FLOWS__CHECKPOINTING'] = 'True'
    r = XpersistResult(
        CacheStore(),
        serializer='xarray.netcdf',
    )

    @task(target='testing.nc', result=r)
    def xarray_task():
        time.sleep(2)
        return xr.tutorial.open_dataset('rasm').isel(time=0)

    with Flow('test') as flow:
        xarray_task()

    state = flow.run(executor=executor)
    assert not state.is_failed()



# Adapted from https://github.com/explosion/thinc
import sys
import typing

import catalogue

# Use typing_extensions for Python versions < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol


_DIn = typing.TypeVar('_DIn')


class Decorator(Protocol):
    def __call__(self, name: str) -> typing.Callable[[_DIn], _DIn]:
        ...


class registry:

    """xpersist's global registry entrypoint.

    This is used to register serializers and other components that are used by xpersist.
    """

    serializers: Decorator = catalogue.create('xpersist', 'serializers', entry_points=True)
    metadata_store: Decorator = catalogue.create('xpersist', 'metadata_store', entry_points=True)

    @classmethod
    def create(cls, registry_name: str, entry_points: bool = False) -> None:
        """Create a new custom registry."""
        if hasattr(cls, registry_name):
            raise ValueError(f"Registry '{registry_name}' already exists")
        reg: Decorator = catalogue.create('xpersist', registry_name, entry_points=entry_points)
        setattr(cls, registry_name, reg)

    @classmethod
    def has(cls, registry_name: str, func_name: str) -> bool:
        """Check whether a function is available in a registry.

        Parameters
        ----------
        registry_name : str
            The name of the registry to check.
        func_name : str
            The name of the function to check.

        Returns
        -------
        bool
            Whether the function is available in the registry.

        """
        if not hasattr(cls, registry_name):
            return False
        reg = getattr(cls, registry_name)
        return func_name in reg

    @classmethod
    def get(cls, registry_name: str, func_name: str) -> typing.Callable:
        """Get a registered function from a given registry.

        Parameters
        ----------
        registry_name : str
            The name of the registry to get the function from.
        func_name : str
            The name of the function to get.

        Returns
        -------
        func : typing.Callable
            The function from the registry.
        """
        if not hasattr(cls, registry_name):
            raise ValueError(f"Unknown registry: '{registry_name}'")
        reg = getattr(cls, registry_name)
        func = reg.get(func_name)
        if func is None:
            raise ValueError(f"Could not find '{func_name}' in '{registry_name}'")
        return func



import fsspec
import pytest
import xarray as xr

from xpersist import CacheStore
from xpersist.cache import Artifact


@pytest.mark.parametrize('readonly', [True, False])
def test_initialization(tmp_path, readonly):
    store = CacheStore(str(tmp_path), readonly=readonly)
    assert isinstance(store.mapper, fsspec.mapping.FSMap)
    assert isinstance(store.raw_path, str)
    assert store.raw_path == str(tmp_path)
    assert store.readonly == readonly


@pytest.mark.parametrize(
    'key, data, serializer',
    [
        ('bar', 'my_data', 'joblib'),
        ('foo', [1, 3, 4], 'auto'),
        ('test.nc', xr.DataArray([1, 2]).to_dataset(name='sst'), 'xarray.netcdf'),
        ('my_dataset.zarr', xr.DataArray([1, 2]).to_dataset(name='sst'), 'xarray.zarr'),
    ],
)
def test_put_and_get(tmp_path, key, data, serializer):
    store = CacheStore(str(tmp_path))
    store.put(key=key, value=data, serializer=serializer)
    assert key in store.keys()
    assert isinstance(store.get_artifact(key), Artifact)
    results = store[key]
    assert results == data


@pytest.mark.parametrize(
    'key, data, serializer',
    [
        ('bar', 'my_data', 'joblib'),
        ('foo', [1, 3, 4], 'auto'),
        ('test.nc', xr.DataArray([1, 2]).to_dataset(name='sst'), 'xarray.netcdf'),
    ],
)
def test_delete(tmp_path, key, data, serializer):
    store = CacheStore(str(tmp_path))
    store.put(key=key, value=data, serializer=serializer)
    assert key in store.keys()
    store.delete(key=key, dry_run=True)
    del store[key]
    assert key not in store.keys()



import pytest
import xarray as xr
import xcollection as xc

import xpersist


@pytest.mark.parametrize(
    'value, expected_serializer',
    [
        (['foo', 'bar'], 'joblib'),
        ({'foo': 'bar'}, 'joblib'),
        (xr.DataArray([1, 2]), 'xarray.netcdf'),
        (xr.DataArray([1, 2]).to_dataset(name='test'), 'xarray.netcdf'),
        (
            xc.Collection(
                {
                    'foo': xr.DataArray([1, 2]).to_dataset(name='test'),
                    'bar': xr.DataArray([1, 2]).to_dataset(name='test'),
                }
            ),
            'xcollection',
        ),
    ],
)
def test_default_serializer(value, expected_serializer):
    assert xpersist.pick_serializer(value) == expected_serializer



# -*- coding: utf-8 -*-


import datetime

import xpersist

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'myst_nb',
    'sphinxext.opengraph',
    'sphinx_copybutton',
    'sphinx_comments',
    'sphinxcontrib.autodoc_pydantic',
    'sphinx_inline_tabs',
]

autodoc_member_order = 'groupwise'

# MyST config
myst_enable_extensions = ['amsmath', 'colon_fence', 'deflist', 'html_image']
myst_url_schemes = ['http', 'https', 'mailto']

# sphinx-copybutton configurations
copybutton_prompt_text = r'>>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: '
copybutton_prompt_is_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Autosummary pages will be generated by sphinx-autogen instead of sphinx-build
autosummary_generate = []
autodoc_typehints = 'none'

# Napoleon configurations

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = False

autodoc_pydantic_model_show_json = True
autodoc_pydantic_settings_show_json = False

jupyter_execute_notebooks = 'cache'
execution_timeout = 600
execution_allow_errors = True


# The master toctree document.
master_doc = 'index'

# General information about the project.
current_year = datetime.datetime.now().year
project = u'xpersist'
copyright = f'2020-{current_year}, xpersist developers'
author = u'xpersist developers'


# The short X.Y version.
version = xpersist.__version__.split('+')[0]
# The full version, including alpha/beta/rc tags.
release = xpersist.__version__


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**.ipynb_checkpoints', 'Thumbs.db', '.DS_Store']


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'
html_title = ''

html_context = {
    'github_user': 'NCAR',
    'github_repo': 'xpersist',
    'github_version': 'main',
    'doc_path': 'docs',
}
html_theme_options = dict(
    # analytics_id=''  this is configured in rtfd.io
    # canonical_url="",
)


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = '../_static/images/NSF_4-Color_bitmap_Logo.png'


# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'


# Output file base name for HTML help builder.
htmlhelp_basename = 'xpersistdoc'


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}


latex_documents = [('index', 'xpersist.tex', u'xpersist Documentation', author, 'manual')]

man_pages = [('index', 'xpersist', u'xpersist Documentation', [author], 1)]

texinfo_documents = [
    (
        'index',
        'xpersist',
        u'xpersist Documentation',
        author,
        'xpersist',
        'One line description of project.',
        'Miscellaneous',
    )
]


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'xarray': ('http://xarray.pydata.org/en/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'fsspec': ('https://filesystem-spec.readthedocs.io/en/latest/', None),
}



import os
import typing

os.environ['PREFECT__FLOWS__CHECKPOINTING'] = 'True'

import pendulum
import pydantic
from prefect.engine.result import Result
from slugify import slugify

from ..cache import CacheStore


@pydantic.dataclasses.dataclass
class XpersistResult(Result):
    """A result class used to store the results of a task in a xpersist cache store.

    Parameters
    ----------
    cache_store : :py:class:`xpersist.cache.CacheStore`
        The cache store to use for storing the result.
    serializer : str
        The serializer to use for storing the result.
    serializer_dump_kwargs : dict
        The keyword arguments to pass to the serializer's `dump` method.
    serializer_load_kwargs : dict
        The keyword arguments to pass to the serializer's `load` method.
    kwargs : dict
        Any additional keyword arguments to pass to the `Result` class.
    """

    cache_store: CacheStore
    serializer: str = 'auto'
    serializer_dump_kwargs: typing.Dict[str, typing.Any] = None
    serializer_load_kwargs: typing.Dict[str, typing.Any] = None
    kwargs: typing.Dict[str, typing.Any] = None

    def __post_init_post_parse__(self):
        self.kwargs = self.kwargs or {}
        self._serializer = self.serializer
        super().__init__(**self.kwargs)
        self.serializer = self._serializer
        self.serializer_dump_kwargs = self.serializer_dump_kwargs or {}
        self.serializer_load_kwargs = self.serializer_load_kwargs or {}

    @property
    def default_location(self) -> str:
        return f"prefect-result-{slugify(pendulum.now('utc').isoformat())}"

    def read(self, location: str) -> Result:
        """Reads a result from the cache store and returns the corresponding `Result` instance.

        Parameters
        ----------
        location : str
            the location to read from

        Returns
        -------
        result : Result
            a new result instance with the data represented by the location
        """
        new = self.copy()
        new.location = location

        self.logger.debug('Starting to read result from {}...'.format(location))
        new.value = self.cache_store.get(key=location, load_kwargs=new.serializer_load_kwargs)
        self.logger.debug('Finished reading result from {}...'.format(location))
        return new

    def write(self, value_: typing.Any, **kwargs: typing.Any) -> Result:
        """Writes the result to a location in the cache store and returns a new `Result`
        object with the result's location.

        Parameters
        ----------
        value_ : typing.Any
            the value to write; will then be stored as the `value` attribute
            of the returned `Result` instance
        kwargs : dict
            if provided, will be used to format the location template
            to determine the location to write to

        Returns
        -------
        result : Result
            A new `Result` instance with the location of the written result.

        """

        new = self.format(**kwargs)
        new.value = value_
        assert new.location is not None

        relevant_context_keys = sorted(
            {
                'today',
                'yesterday',
                'tomorrow',
                'flow_name',
                'task_name',
                'map_index',
                'task_full_name',
                'task_slug',
                'task_tags',
                'task_run_name',
                'flow_id',
                'flow_run_id',
                'flow_run_version',
                'flow_run_name',
                'task_id',
                'task_run_id',
                'task_run_version',
            }
        )
        additional_metadata = {key: kwargs.get(key, '') for key in relevant_context_keys}

        self.logger.debug('Starting to upload result to {}...'.format(new.location))
        self.cache_store.put(
            key=new.location,
            value=new.value,
            serializer=self.serializer,
            dump_kwargs=new.serializer_dump_kwargs,
            additional_metadata=additional_metadata,
        )
        self.logger.debug('Finished uploading result to {}.'.format(new.location))
        return new

    def exists(self, location: str, **kwargs: typing.Any) -> bool:
        """Checks whether the target result exists in the cache store.

        Does not validate whether the result is `valid`, only that it is present.

        Parameters
        ----------
        location : str
            Location of the result in the specific result target.
            Will check whether the provided location exists
        kwargs : dict
            string format arguments for `location`

        Returns
        -------
        _ : bool
            whether or not the target result exists
        """
        return location.format(**kwargs) in self.cache_store


# Fixes https://github.com/samuelcolvin/pydantic/issues/704
XpersistResult.__pydantic_model__.update_forward_refs()



#!/usr/bin/env python3
# flake8: noqa
""" Top-level module for xpersist. """
from pkg_resources import DistributionNotFound, get_distribution

from .cache import CacheStore
from .prefect.result import XpersistResult
from .registry import registry
from .serializers import pick_serializer

try:
    __version__ = get_distribution('xpersist').version
except DistributionNotFound:  # pragma: no cover
    __version__ = 'unknown'  # pragma: no cover



import datetime
import enum
import json
import tempfile
import typing

import fsspec
import pydantic

from .registry import registry
from .serializers import pick_serializer


class Artifact(pydantic.BaseModel):
    """A pydantic model for representing an artifact in the cache."""

    key: str
    serializer: str
    load_kwargs: typing.Optional[typing.Dict] = pydantic.Field(default_factory=dict)
    dump_kwargs: typing.Optional[typing.Dict] = pydantic.Field(default_factory=dict)
    additional_metadata: typing.Optional[typing.Dict] = pydantic.Field(default_factory=dict)
    created_at: typing.Optional[datetime.datetime] = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )
    _value: typing.Any = pydantic.PrivateAttr(default=None)

    class Config:
        validate_assignment = True


class DuplicateKeyEnum(str, enum.Enum):
    skip = 'skip'
    overwrite = 'overwrite'
    check_collision = 'check_collision'
    raise_error = 'raise_error'


@pydantic.dataclasses.dataclass
class CacheStore:
    """Implements caching functionality. Support fsspec backends (local, s3fs, etc...).

    Some backends may require other dependencies. For example to work with S3 cache store,
    s3fs is required.

    Parameters
    ----------
    path : str
        The path to the cache store. This can be a local directory or a cloud storage bucket.
        By default, the path is set to the temporary local directory.
    storage_options : dict
        fsspec parameters passed to the backend file-system such as Google Cloud Storage,
        Amazon Web Service S3.
    readonly : bool
        if True, the cache store is readonly. If False, the cache store is writable.
    on_duplicate_key : DuplicateKeyEnum
        The behavior when a key is duplicated in the cache store. Valid options are:

        - 'skip' (default): do nothing
        - 'overwrite': overwrite the existing artifact
        - 'check_collision': check if the key is already in the cache store and raise an error
        - 'raise_error': raise an error if the key is already in the cache store
    """

    path: str = tempfile.gettempdir()
    readonly: bool = False
    on_duplicate_key: DuplicateKeyEnum = 'skip'
    storage_options: typing.Dict[typing.Any, typing.Any] = None

    def __post_init_post_parse__(self):
        self.storage_options = {} if self.storage_options is None else self.storage_options
        self.mapper = fsspec.get_mapper(self.path, **self.storage_options)
        self.fs = self.mapper.fs
        self.raw_path = self.fs._strip_protocol(self.path)
        self.protocol = self.fs.protocol
        self._ensure_dir(self.raw_path)
        self._suffix = '.artifact.json'
        self._metadata_store_prefix = 'xpersist_metadata_store'
        self._metadata_store_path = self._construct_item_path(self._metadata_store_prefix)
        self._ensure_dir(self._metadata_store_path)

    def _ensure_dir(self, key: str) -> None:
        if not self.fs.exists(key):
            self.fs.makedirs(key, exist_ok=True)

    def _construct_item_path(self, key) -> str:
        return f'{self.path}/{key}'

    def _artifact_meta_relative_path(self, key: str) -> str:
        return f'{self._metadata_store_prefix}/{key}{self._suffix}'

    def _artifact_meta_full_path(self, key: str) -> str:
        return f'{self._metadata_store_path}/{key}{self._suffix}'

    def __contains__(self, key: str) -> bool:
        """Returns True if the key is in the cache store."""
        return self._artifact_meta_relative_path(key) in self.mapper

    def keys(self) -> typing.List[str]:
        """Returns a list of keys in the cache store."""
        keys = self.fs.ls(self._metadata_store_path)
        return [
            key.split(f'{self._metadata_store_prefix}/')[-1].split(self._suffix)[0] for key in keys
        ]

    def delete(self, key: str, dry_run: bool = True) -> None:
        """Deletes the key and corresponding artifact from the cache store.

        Parameters
        ----------
        key : str
            Key to delete from the cache store.
        dry_run : bool
            If True, the key is not deleted from the cache store. This is useful for debugging.
        """
        keys = [key, self._artifact_meta_relative_path(key)]
        if not dry_run:
            self.mapper.delitems(keys)
        else:
            print(f'DRY RUN: would delete items with keys: {repr(keys)}')

    def __getitem__(self, key: str) -> typing.Any:
        """Returns the artifact corresponding to the key."""
        return self.get(key)

    def __setitem__(self, key: str, value: typing.Any) -> None:
        """Sets the key and corresponding artifact in the cache store."""
        self.put(key, value)

    def __delitem__(self, key: str) -> None:
        """Deletes the key and corresponding artifact from the cache store."""
        self.delete(key, dry_run=False)

    @pydantic.validate_arguments
    def get_artifact(self, key: str) -> Artifact:
        """Returns the artifact corresponding to the key.

        Parameters
        ----------
        key : str
            Key to get from the cache store.

        Returns
        -------
        artifact: Artifact
            The artifact corresponding to the key.

        Raises
        ------
        KeyError
            If the key is not in the cache store.

        """
        metadata_file = self._artifact_meta_relative_path(key)
        message = f'{key} not found in cache store: {self._metadata_store_path}'
        if key not in self:
            raise KeyError(message)
        try:
            return Artifact(**json.loads(self.mapper[metadata_file]))
        except Exception as exc:
            raise KeyError(
                f'Unable to load artifact sidecar file {metadata_file} for key: {key}'
            ) from exc

    @pydantic.validate_arguments
    def get(
        self,
        key: str,
        serializer: str = None,
        load_kwargs: typing.Dict[typing.Any, typing.Any] = None,
    ) -> typing.Any:
        """Returns the value for the key if the key is in the cache store.

        Parameters
        ----------
        key : str
            Key to get from the cache store.
        serializer : str
            The name of the serializer you want to use. The built-in
            serializers are:

            - 'auto' (default): automatically choose the serializer based on the type of the value
            - 'xarray.netcdf': requires xarray and netCDF4
            - 'xarray.zarr': requires xarray and zarr

            You can also register your own serializer via the @xpersist.registry.serializers.register decorator.
        load_kwargs : dict
            Additional keyword arguments to pass to the serializer when loading artifact from the cache store.

        Returns
        -------
        value :
            the value for the key if the key is in the cache store.

        Examples
        --------
        >>> from xpersist import CacheStore
        >>> store = CacheStore("/tmp/my-cache")
        >>> store.keys()
        ['foo']
        >>> store.get("foo")
        [1, 2, 3]
        """

        artifact = self.get_artifact(key)
        try:
            serializer_name = serializer or artifact.serializer
            load_kwargs = load_kwargs or artifact.load_kwargs
            serializer = registry.serializers.get(serializer_name)()
            return serializer.load(self._construct_item_path(artifact.key), **load_kwargs)
        except Exception as exc:
            raise ValueError(f'Unable to load artifact {artifact.key} from cache store') from exc

    @pydantic.validate_arguments
    def put(
        self,
        key: str,
        value: typing.Any,
        serializer: str = 'auto',
        dump_kwargs: typing.Dict[typing.Any, typing.Any] = None,
        additional_metadata: typing.Dict[typing.Any, typing.Any] = None,
    ) -> Artifact:
        """Records and serializes key with its corresponding value in the cache store.

        Parameters
        ----------
        key : str
            Key to put in the cache store.
        value : typing.Any
            Value to put in the cache store.
        serializer : str
            The name of the serializer you want to use. The built-in
            serializers are:

            - 'auto' (default): automatically choose the serializer based on the type of the value
            - 'xarray.netcdf': requires xarray and netCDF4
            - 'xarray.zarr': requires xarray and zarr

            You can also register your own serializer via the @xpersist.registry.serializers.register decorator.
        dump_kwargs : dict
            Additional keyword arguments to pass to the serializer when dumping artifact to the cache store.
        additional_metadata : dict
            A dict with types that serialize to json. These fields can be used for searching artifacts in the metadata store.

        Returns
        -------
        value : typing.Any
            Reference to the value that was put in the cache store.

        Examples
        --------
        >>> from xpersist import CacheStore
        >>> store = CacheStore("/tmp/my-cache")
        >>> store.keys()
        []
        >>> store.put("foo", [1, 2, 3])
        >>> store.keys()
        ['foo']

        """
        dump_kwargs = dump_kwargs or {}
        additional_metadata = additional_metadata or {}
        if not self.readonly:
            method = getattr(self, f'_put_{self.on_duplicate_key.value}')
            serializer_name = pick_serializer(value) if serializer == 'auto' else serializer
            artifact = Artifact(
                key=key,
                serializer=serializer_name,
                dump_kwargs=dump_kwargs,
                additional_metadata=additional_metadata,
            )
            artifact._value = value
            method(artifact)
            return artifact._value

    def _put_skip(self, artifact: Artifact) -> None:
        if self._artifact_meta_relative_path(artifact.key) not in self:
            self._put_overwrite(artifact)

    def _put_overwrite(self, artifact: Artifact) -> None:
        serializer = registry.serializers.get(artifact.serializer)()
        with self.fs.transaction:
            serializer.dump(
                artifact._value, self._construct_item_path(artifact.key), **artifact.dump_kwargs
            )

            with self.fs.open(self._artifact_meta_full_path(artifact.key), 'w') as fobj:
                fobj.write(artifact.json(indent=2))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""


from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().strip().split('\n')


setup(
    maintainer='Xdev',
    maintainer_email='xdev@ucar.edu',
    description='xpersist provides custom caching utility functions in Python',
    install_requires=requirements,
    license='Apache Software License 2.0',
    long_description_content_type='text/markdown',
    long_description=long_description,
    name='xpersist',
    packages=['xpersist'],
    url='https://github.com/NCAR/xpersist',
    project_urls={
        'Documentation': 'https://github.com/NCAR/xpersist',
        'Source': 'https://github.com/NCAR/xpersist',
        'Tracker': 'https://github.com/NCAR/xpersist/issues',
    },
    use_scm_version={'version_scheme': 'post-release', 'local_scheme': 'dirty-tag'},
    zip_safe=False,
)



import functools
import typing

import joblib as _joblib
import pydantic
import xarray as xr
import xcollection as xc

from .registry import registry


class Serializer(pydantic.BaseModel):
    """Pydantic model for defining a serializer."""

    name: str
    load: typing.Callable
    dump: typing.Callable


@registry.serializers.register('xarray.zarr')
def xarray_zarr() -> Serializer:
    return Serializer(name='xarray.zarr', load=xr.open_zarr, dump=xr.backends.api.to_zarr)


@registry.serializers.register('xarray.netcdf')
def xarray_netcdf() -> Serializer:
    return Serializer(name='xarray.netcdf', load=xr.open_dataset, dump=xr.backends.api.to_netcdf)


@registry.serializers.register('xcollection')
def xcollection() -> Serializer:
    return Serializer(name='xcollection', load=xc.open_collection, dump=xc.Collection.to_zarr)


@registry.serializers.register('joblib')
def joblib() -> Serializer:
    return Serializer(name='joblib', load=_joblib.load, dump=_joblib.dump)


@functools.singledispatch
def pick_serializer(obj) -> str:
    """Returns the id of the appropriate serializer

    Parameters
    ----------
    obj: any Python object

    Returns
    -------
    id : str
       Id of the serializer
    """

    return registry.serializers.get('joblib')().name


@pick_serializer.register(xr.Dataset)
def _(obj):
    return registry.serializers.get('xarray.netcdf')().name


@pick_serializer.register(xr.DataArray)
def _(obj):
    return registry.serializers.get('xarray.netcdf')().name


@pick_serializer.register(xc.Collection)
def _(obj):
    return registry.serializers.get('xcollection')().name
